Squeeze = Object:extend()

function Squeeze:new(x, y, attributes)
  
  self.type = "squeeze"
  
  if x then
    self.points = {Vec2(x, y)}
  end
  
  self.attributes = attributes or {}
  
  createAttribute(self.attributes, "minSpeed", 10)
  createAttribute(self.attributes, "requiresFacing", true)
  createAttribute(self.attributes, "interactRadius", 100)
  createAttribute(self.attributes, "speedEnterRadius", 20)
  
  
  table.insert(drawTable, self)
  
end




function Squeeze:addPoint(x, y, i)
  if i then
    table.insert(self.points, i, Vec2(x,y))
  else
    table.insert(self.points, Vec2(x,y))
  end
end


function Squeeze:interact(playerPosition)
  if player1.wetness > 0 then
    self:startSqueeze()
  else
    cout("Player is not wet enough!")
  end
end


function Squeeze:getLength()
  
  local length = 0
  
  for i, point in ipairs(self.points) do
    if i ~= 1 then
      local lastPoint = self.points[i-1]
      length = length + vector.getMagnitude(lastPoint - point)
    end
  end
  
  return length
  
end


function Squeeze:getPositionFromDistance(distance)
  local id = distance
  
  if distance < 0 then
    return nil
  end
  
  for i, point in ipairs(self.points) do
    if i ~= 1 then
      local lastPoint = self.points[i-1]
      local length = vector.getMagnitude(lastPoint - point)
      
      if length < distance then
        distance = distance - length
      else
        
        return vector.interpolate(lastPoint, point, distance/length), vector.normalize(point - lastPoint)
      end
      
    end
  end
  
  return nil
  
end


function Squeeze:startSqueeze()
  if not player1.animationInfo.isSqueezing then
    local sideIsFirst
    do
      local playerPos = Vec2(player1.body:getPosition())
      local displacement1 = self.points[1] - playerPos
      local displacement2 = self.points[#self.points] - playerPos
      local distance1 = vector.getMagnitudeSqrd(displacement1) 
      local distance2 = vector.getMagnitudeSqrd(displacement2) 
      
      sideIsFirst = distance1 < distance2 
      
    end

    local speed = math.max(self.attributes.minSpeed*love.physics.getMeter(), vector.getMagnitude(velocityVector))
    
    if sideIsFirst then
      player1.squeezeDistance = 0
      player1.squeezeSpeed = speed
    else
      player1.squeezeDistance = self:getLength()
      player1.squeezeSpeed = -speed
    end
    player1.animationInfo.isSqueezing = true
    player1.sqeezeTrack = self
  end
end




function Squeeze:removePoint(i)
  
  if i then
    table.remove(self.points, i, Vec2(x,y))
  else
    table.remove(self.points, #self.points, Vec2(x,y))
  end
  
end

local radius = 10

function Squeeze:checkSelect()
  
  local mousePos = Vec2(mouse.worldX, mouse.worldY)
  
  
  for i, point in ipairs(self.points) do
    if vector.getMagnitudeSqrd(mousePos - point) < radius^2 then
      return true
    end
  end
  return false
end


function Squeeze:isInteractable()
  
  return true
  
end


function Squeeze:delete()
  self.remove = true
end



function Squeeze:updateBuilder(dt)
  
  
  if not builder.selectedObject then
    if love.keyboard.isDown("delete") then
      if self:checkSelect() then
        
        self:delete()
        
      end
    end
  end
end


function Squeeze:update(dt)
  
  if vector.getMagnitudeSqrd(velocityVector) > (self.attributes.minSpeed*love.physics.getMeter())^2 and player1.wetness > 0 then
    
    local playerPos = Vec2(player1.body:getPosition())
    
    
    
    local displacement1 = self.points[1] - playerPos
    local distance1 = vector.getMagnitudeSqrd(displacement1) 
    
    if distance1 < self.attributes.speedEnterRadius^2 then
      
      if vector.dotProduct(player1.directionVector, displacement1/math.sqrt(distance1)) > 0 then
        self:startSqueeze()
      end
      
    else
    
    
      local displacement2 = self.points[#self.points] - playerPos
      local distance2 = vector.getMagnitudeSqrd(displacement2) 
      if distance2 < self.attributes.speedEnterRadius^2 then
        
        if vector.dotProduct(player1.directionVector, displacement2/math.sqrt(distance2)) > 0 then
          self:startSqueeze()
        end
        
      end
    end
  end
  
end


function Squeeze:drawOverlay()
  
  if debugMode or builderMode then
  
    if self.selected then
      love.graphics.setColor(1,1,1)
      love.graphics.setLineWidth(5)
    else
      love.graphics.setColor(0.5,0.5,1)
      love.graphics.setLineWidth(2.5)
    end
    
    for i, point in ipairs(self.points) do
      if i ~= 1 then
        local lastPoint = self.points[i-1]
        blue.graphics.line(point.x, point.y, lastPoint.x, lastPoint.y)
      end
    end
    
    if self.finalised  then
      love.graphics.setColor(0,0,1)
    else
      love.graphics.setColor(1,0,1)
    end
    for i, point in ipairs(self.points) do
      blue.graphics.circle("line", point.x, point.y, radius)
    end
    love.graphics.setLineWidth(1)
    love.graphics.setColor(1,1,0)
    blue.graphics.circle("line", self.points[1].x, self.points[1].y, self.attributes.speedEnterRadius)
    blue.graphics.circle("line", self.points[#self.points].x, self.points[#self.points].y, self.attributes.speedEnterRadius)
    love.graphics.setColor(1,0,0)
    blue.graphics.circle("line", self.points[1].x, self.points[1].y, self.attributes.interactRadius)
    blue.graphics.circle("line", self.points[#self.points].x, self.points[#self.points].y, self.attributes.interactRadius)
  end
  
  
  Prop.drawInteractPrompt(self)
end