function newTexturedMesh(texture, vertices, scale, offsetX, offsetY)
  
  offsetX = offsetX or 0
  offsetY = offsetY or 0

  local newVertices = {}
  
  
  local textureWidth  = texture:getWidth()
  local textureHeight = texture:getHeight()
  
  local widthMult  = 1/(scale*textureWidth)
  local heightMult = 1/(scale*textureHeight)
  
  
  
  for i = 1, #vertices, 2 do
    
    table.insert(newVertices, {
        vertices[i], 
        vertices[i+1], 
        vertices[i] * widthMult + offsetX, 
        vertices[i+1] * heightMult+offsetY})
  end
  


  if #newVertices > 4 then
    local mesh = love.graphics.newMesh(newVertices)
    mesh:setTexture(texture)
    return mesh
  else
    cout("Invalid polygon")
    return nil
    --error("Not enough vertices to create mesh ("..#newVertices..")")
  end
  
end



local function getPolygonNormals(polygon)
  
  local normals = {}
  
  for i = 1, #polygon, 2 do


    local x1 = polygon[i]
    local y1 = polygon[i+1]
    local x2, y2
    
    if i == #polygon-1 then
      x2 = polygon[1]
      y2 = polygon[2]
    else
      x2 = polygon[i+2]
      y2 = polygon[i+3]
    end
    
    
    local vx = x2 - x1
    local vy = y2 - y1
    
    
    local length = (vx^2 + vy^2)^0.5
    
    
    local nx
    local ny
    
    if getPolygonArea(polygon) > 0 then
      nx = vy/length*100
      ny = -vx/length*100
    else
      nx = -vy/length*100
      ny = vx/length*100
    end
    
    table.insert(normals, nx)
    table.insert(normals, ny)
  end
  
  return normals
end


local function bresenhamLine(dx, dy)

  local newLine = {}
  local x = 0
  local y = 0
  
  local pk = 2 * dy - dx
  
  while x < dx do
    if pk < 0 then
      pk = pk + 2 * dy
      x = x + 1
      y = y  
    else
      
      y = y + 1
      pk = pk + 2*dy - 2*dx
      x = x + 1
    end
    table.insert(newLine, x)
    table.insert(newLine, y)
  end

  return newLine
end



local function pixellateLine(pixelSize, startX, startY, endX, endY)
  
  local flipX = 1
  local flipY = 1
  local swapAxes = false
  
  local offsetX = roundToNearest(startX, pixelSize) - startX 
  local offsetY = roundToNearest(startY, pixelSize) - startY
  
  
  
  startX = roundToNearest(startX, pixelSize)
  startY = roundToNearest(startY, pixelSize)
  
  endX = roundToNearest(endX, pixelSize)
  endY = roundToNearest(endY, pixelSize)

  if startX >= endX then
    flipX = -1
  end
  if startY >= endY then
    flipY = -1
  end
  
  
  if math.abs(endX - startX) < math.abs(endY - startY) then
    local temp = startX
    startX = startY
    startY = temp
    
    temp = endX
    endX = endY
    endY = temp
    
    swapAxes = true
  end
  
  
  startX = (startX/pixelSize)
  startY = (startY/pixelSize)
  
  endX = (endX/pixelSize)
  endY = (endY/pixelSize)

  local line = bresenhamLine(math.abs(math.floor(endX-startX)), math.abs(math.floor(endY-startY)))
  
  for i, v in ipairs(line) do
    if i % 2 == 1 then
      if swapAxes then
        line[i] = (line[i]*flipY + startX) * pixelSize
      else
        line[i] = (line[i]*flipX + startX) * pixelSize
      end
    else
      if swapAxes then
        line[i] = (line[i]*flipX + startY) * pixelSize
      else
        line[i] = (line[i]*flipY + startY) * pixelSize
      end
    end
  end
  
  
  if swapAxes then
    for i = 1, #line, 2 do
      local temp = line[i]
      line[i] = line[i+1]
      line[i+1] = temp
    end
  end

  return line
end




local function createPixellatedPolygon(polygon, pixelSize)
  
  polygon = copyTable(polygon)
  
  for i, v in ipairs(polygon) do
    polygon[i] = polygon[i] + pixelSize/2
  end
  
  
  local newPolygon = {}
  local normals = getPolygonNormals(polygon)
  local outlinePolygon = {}
  
  
  for i = 1, #polygon, 2 do
    local x1
    local y1
    local x2
    local y2
    x2 = polygon[i]
    y2 = polygon[i+1]
    
    if i == 1 then
      x1 = polygon[#polygon-1]
      y1 = polygon[#polygon]
    else
      x1 = polygon[i-2]
      y1 = polygon[i-1]
    end
    
    local nx, ny
    if i == 1 then
      nx = normals[#normals-1]
      ny = normals[#normals]
      
    else
      nx = normals[i-2]
      ny = normals[i-1]
    end
    
    if i-4 < 1 then
      n2x = normals[#normals + i-4]
      n2y = normals[#normals + i-3]
    else
      n2x = normals[i-4]
      n2y = normals[i-3]
    end
    
    
    
    local line = pixellateLine(pixelSize, x1, y1, x2, y2)

    local tmpLine = copyTable(line)
    
    table.insert(tmpLine, 1, roundToNearest(x1, pixelSize))
    table.insert(tmpLine, 2, roundToNearest(y1, pixelSize))
    
    
    local oldPolygonLength = #outlinePolygon
    
    local lineLength = #tmpLine
    
    for i = 1, lineLength, 2 do
      
      local line = tmpLine
      
      local x1 = tmpLine[i]
      local y1 = tmpLine[i+1]
      
      local oldX1 = x1
      local oldY1 = y1
      
      x2 = line[i+2]
      y2 = line[i+3]
      
      local dx, dy
      
      if x2 then
        dx = x2 - x1
        dy = y2 - y1
      end
      if nx > 0 then
        x1 = x1 + pixelSize/2
      else
        x1 = x1 - pixelSize/2
      end
      
      if ny > 0 then
        y1 = y1 + pixelSize/2
      else
        y1 = y1 - pixelSize/2
      end
      
      if  i == 1 then
        if not (outlinePolygon[#outlinePolygon-1] == x1 or outlinePolygon[#outlinePolygon] == y1) then
          
          local anx
          local any
          
          if nx + n2x ~= 0 then
            anx = (nx + n2x)/math.abs(nx + n2x)
          else
            anx = 1
          end
          
          if ny + n2y ~= 0 then
            any = (ny + n2y)/math.abs(ny + n2y)
          else
            any = 1
          end 
          
          table.insert(outlinePolygon, oldX1 + pixelSize*anx/2)
          table.insert(outlinePolygon, oldY1 + pixelSize*any/2)
        end
      end
        
      table.insert(outlinePolygon, x1)
      table.insert(outlinePolygon, y1)
      
      -- Add the corners
      if dx ~= 0 and dy ~= 0 and i ~= lineLength-1 then
        -- If a corner should be added, add it
        if getPolygonArea(polygon) > 1 then
        
          if nx > 0 and ny > 0 then
            table.insert(outlinePolygon, x1-pixelSize)
            table.insert(outlinePolygon, y1)
          elseif nx < 0 and ny < 0 then
            table.insert(outlinePolygon, x1+pixelSize)
            table.insert(outlinePolygon, y1)
          elseif nx < 0 and ny > 0 then
            table.insert(outlinePolygon, x1)
            table.insert(outlinePolygon, y1-pixelSize)
            
          elseif nx > 0 and ny < 0 then
            table.insert(outlinePolygon, x1)
            table.insert(outlinePolygon, y1+pixelSize)
          end 
        else
          if nx > 0 and ny > 0 then
            table.insert(outlinePolygon, x1)
            table.insert(outlinePolygon, y1-pixelSize)
          elseif nx < 0 and ny < 0 then
            table.insert(outlinePolygon, x1)
            table.insert(outlinePolygon, y1+pixelSize)
          elseif nx < 0 and ny > 0 then
            table.insert(outlinePolygon, x1+pixelSize)
            table.insert(outlinePolygon, y1)
          elseif nx > 0 and ny < 0 then
            table.insert(outlinePolygon, x1-pixelSize)
            table.insert(outlinePolygon, y1)
          end
        end
      end
    end
    
    
    
    for i, v in ipairs(line) do
      table.insert(newPolygon, v)
    end
  end
  

  for i = 1, #outlinePolygon, 2 do
    local x1 = outlinePolygon[i]
    local y1 = outlinePolygon[i+1]
    local x2
    local y2
    
    
    x2 = outlinePolygon[(i+3) % #outlinePolygon+1]
    y2 = outlinePolygon[(i+4) % #outlinePolygon+1]
    
    
    if x1 == x2 and y1 == y2 then

      local length = #outlinePolygon
      table.remove(outlinePolygon, (i+2) % length)
      table.remove(outlinePolygon, (i+2) % length)
      i = i + 2
    end
  end
  
  
  for i = 1, #outlinePolygon, 2 do
    local x1
    local y1
    local x2
    local y2
    x2 = outlinePolygon[i]
    y2 = outlinePolygon[i+1]
    
    if i == 1 then
      x1 = outlinePolygon[#outlinePolygon-1]
      y1 = outlinePolygon[#outlinePolygon]
    else
      x1 = outlinePolygon[i-2]
      y1 = outlinePolygon[i-1]
    end
    
    if x1 == x2 and y1 == y2 then
      table.remove(outlinePolygon, i)
      table.remove(outlinePolygon, i)
      i = i - 2
    end
  end
  
  
  -- Need to make sure the polygon won't crash the game by having overlapping vertices
  
  local i = 1
  while i < #outlinePolygon do
    local x1 = outlinePolygon[i]
    local y1 = outlinePolygon[i+1]
    
    for j = #outlinePolygon, 1, -2 do
      
      local x2 = outlinePolygon[j-1]
      local y2 = outlinePolygon[j]
      
      if x1 == x2 and y1 == y2 then
        
        for k = i, j-2, 2 do
          
          table.remove(outlinePolygon, i+1)
          table.remove(outlinePolygon, i+1)
          
        end
      end
    end
  i = i + 2
  end
  
  
  --Remove unneeded vertices
  
  local offset = 0
  
  for i = 1, #outlinePolygon, 2 do
    
    local j = i - offset
    local x1 = outlinePolygon[j]
    local y1 = outlinePolygon[j+1]
    local x2 = outlinePolygon[(j+3) % #outlinePolygon+1]
    local y2 = outlinePolygon[(j+4) % #outlinePolygon+1]
    
    
    if x1 - x2 == 0 or y1 - y2 == 0 then
      local length = #outlinePolygon
      table.remove(outlinePolygon, (j+2) % length)
      table.remove(outlinePolygon, (j+2) % length)
      i = i + 2
      offset = offset + 2
    end
  end
  
  for i, v in ipairs(outlinePolygon) do
    outlinePolygon[i] = outlinePolygon[i] - pixelSize/2
  end
  
  return outlinePolygon
end


local function convertPolygonToTriangles(polygon)
  
  local newPoly = {}
  
  if #polygon >= 6 then

    local tris = love.math.triangulate(polygon)
    
    for i, triangle in ipairs(tris) do
      
      for i, component in ipairs(triangle) do
        table.insert(newPoly, component)
      end
      
    end
  end
  return newPoly
end


function createPixpolMesh(texture, inputPolygon, pixelSize, offsetX, offsetY)
  
  
  if not texture then
    error("No texture given for pixpol mesh!", 2)
  end
  if not inputPolygon then
    error("No input polygon given for pixpol mesh!", 2)
  end
  if not pixelSize then
    error("No pixel size given for pixpol mesh!", 2)
  end
  
  
  local pixpol = convertPolygonToTriangles(createPixellatedPolygon(inputPolygon, pixelSize))
  
  local mesh = newTexturedMesh(texture, pixpol, pixelSize, offsetX, offsetY)
  
  if mesh then
    mesh:setDrawMode("triangles")
  end
  return mesh
end


function roundToNearest(value, precision)
  precision = precision or 1
  return math.floor(value/precision+0.5)*precision
end